﻿using System;
using System.Drawing;
using System.Globalization;
using System.IO;
using System.Collections.Generic;
using System.Web;
using System.Web.UI;
using System.Web.UI.WebControls;
using C1.Web.UI.Controls.C1GridView;

/// <summary>
/// Summary description for C1GridViewExport
/// </summary>
public static class C1GridViewExport
{
    //---------------------------------------------------------------------------------------------------
    #region ** PDF export

    static Font _bodyFont = new Font("Georgia", 10);
    static Font _headerFont = new Font("Georgia", 10, FontStyle.Bold);
    static Font _symbolFont = new Font("WingDings", 10, FontStyle.Bold);

    const string CHKSTR_CHECKED = "\xf0fe";
    const string CHKSTR_UNCHECKED = "\xf0a8";

    // export a C1GridView to pdf
    public static bool ExportPdf(Page page, C1GridView grid)
    {
        // get pdf stream
        var stream = GetPdfStream(grid);

        // no stream? we're done
        if (stream == null || stream.Length == 0)
        {
            return false;
        }

        // copy stream to Page's Response object
        WriteStreamToPage(page, stream, "application/pdf");

        // done
        return true;
    }

    // create a Pdf stream with the grid contents
    static MemoryStream GetPdfStream(C1GridView grid)
    {
        int[] mapping;
        List<C1BaseField> columns = GetVisibleColumnsWithMapping(grid, out mapping);
        
        // make sure grid has at least one visible column  
        if (columns.Count == 0)
        {
            return null;
        }

        // start with new empty document
        var doc = new C1.C1Pdf.C1PdfDocument();

        // get render rectangle (1-inch margin all around)
        var rc = GetPageRectangle(doc);

        // calculate column widths
        int cellOffset = grid.RowHeader.Visible ? 1 : 0;
        var widths = GetColumnWidths(doc, grid, rc, cellOffset, columns, mapping);

        // render the grid header
        if (grid.ShowHeader)
        {
            var row = grid.HeaderRows[grid.HeaderRows.Length - 1];
            rc = RenderGridRow(doc, rc, grid, row, widths, true, cellOffset, columns, mapping);
        }

        // render the grid body
        foreach (C1GridViewRow row in grid.Rows)
        {
            rc = RenderGridRow(doc, rc, grid, row, widths, false, cellOffset, columns, mapping);
        }

        // create output stream
        var ms = new MemoryStream();
        doc.Save(ms);
        return ms;
    }

    // get page rectangle (one inch margins)
    static RectangleF GetPageRectangle(C1.C1Pdf.C1PdfDocument doc)
    {
        var rcPage = doc.PageRectangle;
        rcPage.Inflate(-72, -72);
        return rcPage;
    }

    // calculate column widths to fit the page
    static float[] GetColumnWidths(C1.C1Pdf.C1PdfDocument doc, C1GridView grid, RectangleF rc,
           int cellOffset, List<C1BaseField> columns, int[] mapping)
    {
        // dimension column width vector
        var widths = new float[columns.Count];

        // measure header cells
        if (grid.ShowHeader)
        {
            var lastHeaderRow = grid.HeaderRows[grid.HeaderRows.Length - 1];
            for (int col = 0; col < columns.Count; col++)
            {
                string text = HttpUtility.HtmlDecode(columns[col].HeaderText);
                var width = doc.MeasureString(text, _headerFont).Width;
                widths[col] = width;
            }
        }

        // measure body cells
        foreach (C1GridViewRow row in grid.Rows)
        {
            for (int col = 0; col < columns.Count; col++)
            {
                string text = HttpUtility.HtmlDecode(row.Cells[mapping[col] + cellOffset].Text);
                var width = doc.MeasureString(text, _bodyFont).Width;
                widths[col] = Math.Max(widths[col], width);
            }
        }

        // adjust to fit the page
        float totalWidth = 0;
        for (int col = 0; col < widths.Length; col++)
        {
            totalWidth += widths[col];
        }
        var adjustment = rc.Width / totalWidth;
        if (adjustment < 1)
        {
            for (int col = 0; col < widths.Length; col++)
            {
                widths[col] *= adjustment;
            }
        }

        // done
        return widths;
    }

    // render a grid row
    static RectangleF RenderGridRow(C1.C1Pdf.C1PdfDocument doc, RectangleF rc, C1GridView grid,
        C1GridViewRow row, float[] widths, bool header, int cellOffset, List<C1BaseField> columns, int[] mapping)
    {
        const int CELL_MARGIN = 4;

        // get row cells
        var cells = row.Cells;

        // calculate cell rectangle
        RectangleF rcCell = rc;
        rcCell.Height = 0;

        // calculate cell height (max of all columns)
        var font = header ? _headerFont : _bodyFont;
        for (int col = 0; col < columns.Count; col++)
        {
            rcCell.Width = widths[col];

            string text = (header)
                ? HttpUtility.HtmlDecode(columns[col].HeaderText)
                : HttpUtility.HtmlDecode(cells[mapping[col] + cellOffset].Text);

            rcCell.Inflate(-CELL_MARGIN, 0);
            float height = doc.MeasureString(text, font, rcCell.Width).Height;
            rcCell.Inflate(CELL_MARGIN, 0);
            rcCell.Height = Math.Max(rcCell.Height, height);
        }

        // break page if we have to
        var rcPage = GetPageRectangle(doc);
        if (!header && rcCell.Bottom > rcPage.Bottom)
        {
            doc.NewPage();
            rc = rcPage;
            if (grid.ShowHeader)
            {
                var lastHeaderRow = grid.HeaderRows[grid.HeaderRows.Length - 1];
                rc = RenderGridRow(doc, rc, grid, lastHeaderRow, widths, true, cellOffset, columns, mapping);
            }
            rcCell.Y = rc.Y;
        }

        // center cell content vertically
        var sf = new StringFormat();
        sf.LineAlignment = StringAlignment.Center;

        // render data cells
        using (Pen pen = new Pen(Brushes.Gray, 0.1f))
        {
            for (int col = 0; col < columns.Count; col++)
            {
                // get font
                font = header ? _headerFont : _bodyFont;

                // get content
                var cell = cells[mapping[col] + cellOffset];
                string text = (header)
                    ? HttpUtility.HtmlDecode(columns[col].HeaderText)
                    : HttpUtility.HtmlDecode(cell.Text);

                // set horizontal alignment
                double d;
                sf.Alignment = (double.TryParse(text, NumberStyles.Any, CultureInfo.CurrentCulture, out d))
                    ? StringAlignment.Far
                    : StringAlignment.Near;

                // handle check boxes
                if (string.IsNullOrEmpty(text) && cell.Controls.Count > 0 && cell.Controls[0] is CheckBox)
                {
                    sf.Alignment = StringAlignment.Center;
                    var cb = cell.Controls[0] as CheckBox;
                    text = cb.Checked ? CHKSTR_CHECKED : CHKSTR_UNCHECKED;
                    font = _symbolFont;
                }

                // render cell
                rcCell.Width = widths[col];
                doc.DrawRectangle(pen, rcCell);
                rcCell.Inflate(-CELL_MARGIN, 0);
                doc.DrawString(text, font, Brushes.Black, rcCell, sf);
                rcCell.Inflate(CELL_MARGIN, 0);
                rcCell.Offset(rcCell.Width, 0);
            }
        }

        // update rectangle and return it
        rc.Offset(0, rcCell.Height);
        return rc;
    }

    #endregion

    //---------------------------------------------------------------------------------------------------
    #region ** Excel export

    // export a C1GridView to an Excel stream
    public static bool ExportExcel(Page page, C1GridView grid)
    {
        // get excel stream
        var stream = GetExcelStream(grid);

        // no stream? we're done
        if (stream == null || stream.Length == 0)
        {
            return false;
        }

        // copy stream to Page's Response object
        WriteStreamToPage(page, stream, "application/vnd.ms-excel");

        // done
        return true;
    }

    // create an Excel stream for a C1GridView
    static MemoryStream GetExcelStream(C1GridView grid)
    {
        int[] mapping;
        List<C1BaseField> columns = GetVisibleColumnsWithMapping(grid, out mapping);

        // make sure grid has at least one visible column
        if (columns.Count == 0)
        {
            return null;
        }

        // start with new empty book
        var book = new C1.C1Excel.C1XLBook();
        var sheet = book.Sheets[0];

        // export header cells
        if (grid.ShowHeader)
        {
            for (int col = 0; col < columns.Count; col++)
            {
                sheet[0, col].Value = columns[col].HeaderText;
            }
        }

        // get row and cell offset
        int rowOffset = grid.ShowHeader ? 1 : 0;
        int cellOffset = grid.RowHeader.Visible ? 1 : 0;

        // export body
        double dbl;
        DateTime dateTime;
        for (int row = 0; row < grid.Rows.Count; row++)
        {
            for (int col = 0; col < columns.Count; col++)
            {
                // get cell and value
                var cell = grid.Rows[row].Cells[mapping[col] + cellOffset];
                string text = HttpUtility.HtmlDecode(cell.Text);

                if (string.IsNullOrEmpty(text) && cell.Controls.Count > 0 && cell.Controls[0] is CheckBox)
                {
                    // boolean value
                    var cb = cell.Controls[0] as CheckBox;
                    sheet[row + rowOffset, col].Value = cb.Checked;
                }
                else if (double.TryParse(text, NumberStyles.Any, CultureInfo.CurrentCulture, out dbl))
                {
                    // numeric value
                    sheet[row + rowOffset, col].Value = dbl;
                }
                else if (DateTime.TryParse(text, CultureInfo.CurrentCulture, DateTimeStyles.None, out dateTime))
                {
                    // date/time value
                    sheet[row + rowOffset, col].Value = dateTime;
                }
                else
                {
                    // everything else is text
                    sheet[row + rowOffset, col].Value = text;
                }
            }
        }

        // freeze header row and give it a background
        if (grid.ShowHeader)
        {
            sheet.Rows.Frozen = 1;
            var style = new C1.C1Excel.XLStyle(book);
            style.BackColor = Color.LightGray;
            sheet.Rows[0].Style = style;
        }

        // create and return stream
        var ms = new MemoryStream();
        book.Save(ms);
        return ms;
    }

    #endregion

    //---------------------------------------------------------------------------------------------------
    #region ** utilities

    // write a stream into the page response object
    static void WriteStreamToPage(Page page, MemoryStream stream, string contentType)
    {
        // get response object, clear it
        var rsp = page.Response;
        rsp.Clear();
        rsp.ClearContent();
        rsp.ClearHeaders();

        // add Accept-Header header (required when https is used)
        string len = stream.Length.ToString();
        rsp.AddHeader("Accept-Header", len);

        // add Content-Length header
        rsp.AddHeader("Content-Length", len);

        // write pdf stream into response buffer
        rsp.ContentType = contentType;
        rsp.OutputStream.Write(stream.GetBuffer(), 0, (int)stream.Length);

        // done
        rsp.Flush();
        rsp.SuppressContent = true;
    }

    private struct ColumnInfo
    {
        public readonly C1BaseField Field;
        public readonly bool IsVisible;

        public ColumnInfo(C1BaseField field, bool isVisible)
        {
            Field = field;
            IsVisible = isVisible;
        }
    }

    private static List<C1BaseField> GetVisibleColumnsWithMapping(C1GridView gridView, out int[] mapping)
    {
        List<ColumnInfo> allColumns = new List<ColumnInfo>();

        GetColumnsR(gridView.Columns, allColumns, true);

        List<C1BaseField> columns = new List<C1BaseField>();
        List<int> mapList = new List<int>();

        for (int i = 0; i < allColumns.Count; i++)
        {
            if (allColumns[i].IsVisible)
            {
                columns.Add(allColumns[i].Field);
                mapList .Add(i);
            }
        }

        mapping = mapList.ToArray();
        return columns;
    }

    private static void GetColumnsR(C1BaseFieldCollection root, List<ColumnInfo> addTo, bool parentVisible)
    {
        foreach (C1BaseField field in root)
        {
            C1Band band = field as C1Band;
            if (band != null)
            {
                if (!band.HasVisibleChild() && band.Visible && parentVisible)
                {
                    addTo.Add(new ColumnInfo(band, true));
                }

                GetColumnsR(band.Columns, addTo, parentVisible && band.Visible);
            }
            else
            {
                addTo.Add(new ColumnInfo(field, field.Visible && parentVisible));
            }
        }
    }

    private static bool HasVisibleChild(this C1Band band)
    {
        foreach (C1BaseField field in band.Columns)
        {
            if (field.Visible)
            {
                return true;
            }
        }

        return false;
    }

    #endregion
}